
local utf8 = require("utf8")

local app = require "src.app"
local oops = require "src.utils.oops"

local Screen = require "src.screens.screen"

local SaveFileScreen = Screen:makeClass
{
    handlers = {},
    filepath = "../../data/newfile.bin"
}

function SaveFileScreen:init(callback)
    Screen.init(self)
    
    if not callback or type(callback) ~= "function" then
        error("Callback function wasn't supplied!")
    end
    
    self.callback = callback
end

function SaveFileScreen.handlers.keypressed(self, key, scancode, isRepeat)
    if key == "backspace" then
        local offset = utf8.offset(self.filepath, -1)
        
        if offset then
            self.filepath = self.filepath:sub(1, offset - 1)
            self.dirty = true
        end
    elseif key == "return" and self.callback then
        app.popScreen()
        self.callback(self.filepath)
    end
end

function SaveFileScreen.handlers.textinput(self, text)
    self.filepath = self.filepath .. text
    self.dirty = true
end

function SaveFileScreen:draw()
    love.graphics.setColor(255, 255, 255)
    
    love.graphics.setFont(app.res.uifont)
    love.graphics.print("Please choose a new filename and press Enter:", 32, 32)
    love.graphics.setFont(app.res.uifontLarge)
    
    local width, height = love.graphics.getDimensions()
    love.graphics.printf(self.filepath, 32, 64, width - 64)
end

return SaveFileScreen
